/*
 *  Project: File Descriptors Table
 *  File: fdt.c
 *  Date: 01/30/2006
 *  Author. R. Gioiosa
 *  Version: 1.2
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/file.h>
#include <linux/version.h>

#define MODULE_VERS "1.1"
#define MODULE_NAME "fdt"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
#define LKH_HAS_FDTABLE
#endif

#define DEBUG_ON

#ifdef DEBUG_ON
#define Dprintk(ARG...) printk(KERN_DEBUG MODULE_NAME ": " ARG)
#else
#define Dprintk(ARG...) 
#endif

static struct proc_dir_entry* fdt_pde;

struct fdt_data_t
{
  pid_t pid;
  struct semaphore fdt_sem;
} fdt_data;

static int fdt_write(struct file* file,const char __user* buffer,unsigned long count,void* data)
{
  const int BLEN = 32;
  char buf[BLEN];
  pid_t pid;

  if(!count)
    {
      Dprintk("Count is 0!!\n");
      return 0;
    }
  if(count>BLEN)
    {
      Dprintk("Count is too large, set to %d\n",BLEN-1);      
      count = BLEN-1;
    }
  if(copy_from_user(buf,buffer,count))
    {
      Dprintk("Error while copying from user space!\n");      
      return -EFAULT;
    }
  buf[count] = '\0';
  if(sscanf(buf,"%d",&pid) != 1)
    {
      Dprintk("The value provided is not a valid number!\n");      
      return -EFAULT;
    }
  Dprintk("writing pid=%d\n",pid);
  down(&fdt_data.fdt_sem);
  fdt_data.pid = pid;
  up(&fdt_data.fdt_sem);
  return count;
}

static int fdt_read(char* page,char** start,off_t off,int count,int* eof,void* data)
{
  int written=0,i=0;
  task_t* tsk;
  struct file** fd = NULL;
  int fmax = 0;

  down(&fdt_data.fdt_sem);
  if((tsk=find_task_by_pid(fdt_data.pid)) == NULL)
    {
      Dprintk("Process %d does not exist!\n",fdt_data.pid);
      up(&fdt_data.fdt_sem);
      return 0;
    }

  /*
   * Since 2.6.14 a new struct (fdtable) has been added in order to implement
   * a new lock-free model of file descriptor management based on RCU. 
   * See Documentation/filesystem/files.txt
   */
  up(&fdt_data.fdt_sem);
#ifdef LKH_HAS_FDTABLE
  fd = tsk->files->fdt->fd;
  fmax = tsk->files->fdt->max_fds;
#else
  fd = tsk->files->fd;
  fmax = tsk->files->max_fds;
#endif
  read_lock(tsk->files->file_lock);
  for(i=0; i<fmax; i++)
    {
      if(fd[i])
	{
	  written += sprintf(page+written,"%d %Lu\n",i,fd[i]->f_pos);
	  Dprintk("fd=%d pos=%Lu\n",i,fd[i]->f_pos);	  
	}
    }
  read_unlock(tsk->files->file_lock);
  Dprintk("reading pid = %d page=%p *start=%p written=%d off=%lu\n",tsk->pid,page,*start,written,off);
  *start = page+off;
  if(strlen(page) == written)
    *eof = 1;
  return written-off;
}

static int __init fdt_init(void)
{
  fdt_pde = create_proc_entry("fdt",0666,NULL);
  if(!create_proc_entry)
    {
      printk("%s: Error allocating proc_dir_entry",MODULE_NAME);
      return -1;
    }
  fdt_pde->owner = THIS_MODULE;
  fdt_pde->read_proc = fdt_read;
  fdt_pde->write_proc = fdt_write;
  fdt_pde->data = NULL;

  fdt_data.pid = 0;
  init_MUTEX(&fdt_data.fdt_sem); 
  Dprintk("File descriptor table loaded!\n");
  return 0;
}

static void __exit fdt_exit(void)
{
  remove_proc_entry("fdt",NULL);
  Dprintk("File descriptor table unloaded!\n");
}

module_init(fdt_init);
module_exit(fdt_exit);

MODULE_AUTHOR("Roberto Gioiosa");
MODULE_DESCRIPTION("File descriptors table");
MODULE_LICENSE("GPL");
MODULE_VERSION(MODULE_VERS);
